from typing import Dict, Set
import os
import locale

import amulet_map_editor
from amulet_map_editor.api import config as CONFIG, log

# there might be a proper way to do this but this should be enough for now

_lang_dirs: Set[str] = set()  # the language directories
_lang: Dict[
    str, str
] = {}  # a storage for the language strings. unique_identifier: language_string

_default_language = "en"

# find out what language the user is using.
try:
    # try getting the OS language
    _language = locale.getdefaultlocale()[0]
except:
    # if that fails use the default language
    _language = _default_language

# if a language is set in the config use that
_language = CONFIG.get("amulet_meta", {}).get("lang", _language)

if _language is None:
    _language = _default_language


def register_lang_directory(lang_dir: str):
    """Register a new language directory.
    Use this, if you are a third party program, to register new translations for your program.

    :param lang_dir: The directory containing language files to register. Files must be of the format <language_code>.lang eg en_US.lang
    :return:
    """
    if os.path.isdir(lang_dir):
        if lang_dir in _lang_dirs:
            log.warning(
                f"The language directory {lang_dir} has already been registered."
            )
        else:
            _lang_dirs.add(lang_dir)
            default_lang_path = os.path.join(lang_dir, f"{_default_language}.lang")
            default_lang = _load_lang_file(default_lang_path)
            if _default_language != _language:
                # if the language set is not the default, load the language
                def _load_lang(lang_name):
                    lang_path = os.path.join(lang_dir, f"{lang_name}.lang")
                    lang = _load_lang_file(lang_path)

                    # sanity check to make sure that all entries in the user language are in the default
                    diff = set(lang).difference(set(default_lang))
                    if diff:
                        print(
                            f"There are {len(diff)} language entries defined in {lang_path} that are not defined in {default_lang_path}\n{diff}"
                        )

                    # merge in the user language
                    default_lang.update(lang)

                if "_" in _language:
                    # first load region independent language files
                    # eg for en_US load en.lang
                    _load_lang(_language.split("_")[0])

                # second load region specific language files
                # eg en_US.lang
                _load_lang(_language)

            # merge the loaded language entries
            for unique_identifier, language_string in default_lang.items():
                if unique_identifier in _lang:
                    log.warning(
                        f"The language entry {unique_identifier} added in {lang_dir} is already used. Please add a unique prefix."
                    )
                else:
                    _lang[unique_identifier] = language_string


def _load_lang_file(lang_path: str) -> Dict[str, str]:
    """Loads a language file and returns the result as a dictionary.
    Will return an empty dictionary if the path does not exist.

    :param lang_path: The lang file path to load.
    :return: A dictionary mapping the string identifier to the string
    """
    lang = {}
    if os.path.isfile(lang_path):
        with open(lang_path, encoding="utf-8") as f:
            for line in f.readlines():
                lstrip_line = line.lstrip()
                if not lstrip_line.startswith("#"):
                    split_line = lstrip_line.split("=", 1)
                    if len(split_line) == 2:
                        unique_identifier = split_line[0].rstrip()
                        language_string = split_line[1].replace("\\n", "\n").strip()
                        lang[unique_identifier] = language_string
    return lang


# load the normal language directory
register_lang_directory(
    os.path.join(os.path.dirname(amulet_map_editor.__file__), "lang")
)


def get(unique_identifier):
    if unique_identifier not in _lang:
        # help debugging referenced lang entries that do not exist
        log.info(f"Could not find lang entry for {unique_identifier}")
    return _lang.get(unique_identifier, unique_identifier)



from OpenGL.GL import (
    glClearColor,
    glBindVertexArray,
    glGenVertexArrays,
    GL_VERTEX_SHADER,
    GL_FRAGMENT_SHADER,
    glGetString,
    GL_SHADING_LANGUAGE_VERSION,
    GL_VERSION,
)
import OpenGL.GL.shaders
import os
from typing import Dict, Tuple, Any
import re

shader_dir = os.path.join(os.path.dirname(__file__))
_shaders: Dict[Tuple[str, str], Any] = {}

GL_VERSION_MATCH = re.compile(r"^(?P<major>\d+)\.(?P<minor>\d+)?")


def get_shader(
    context_identifier: str, shader_name: str
) -> OpenGL.GL.shaders.ShaderProgram:
    shader_key = (context_identifier, shader_name)
    if shader_key not in _shaders:
        gl_version_match = GL_VERSION_MATCH.match(
            glGetString(GL_SHADING_LANGUAGE_VERSION).decode("utf-8")
        )
        if gl_version_match:
            gl_version_tuple = tuple(int(v) for v in gl_version_match.groups())
            if gl_version_tuple >= (3, 30):
                gl_version = "330"  # opengl 3.3
            else:
                gl_version = "120"  # opengl 2.1
        else:
            # in theory this shouldn't happen if the version is correctly formatted.
            gl_version = "330"

        def compile_shader():
            return OpenGL.GL.shaders.compileProgram(
                _load_shader(
                    os.path.join(shader_dir, f"{shader_name}_{gl_version}.vert"),
                    GL_VERTEX_SHADER,
                ),
                _load_shader(
                    os.path.join(shader_dir, f"{shader_name}_{gl_version}.frag"),
                    GL_FRAGMENT_SHADER,
                ),
            )

        try:
            shader = compile_shader()
        except OpenGL.GL.shaders.ShaderValidationError:  # on Mac the above fails if there is no VBO bound
            glBindVertexArray(glGenVertexArrays(1))
            shader = compile_shader()
            glBindVertexArray(0)

        _shaders[shader_key] = shader
    return _shaders[shader_key]


def _load_shader(shader_path, shader_type):
    with open(shader_path, "rb") as f:
        shader = f.read()
    return OpenGL.GL.shaders.compileShader(shader, shader_type)



from typing import Callable, Any, Optional

from .base_operation_loader import BaseOperationLoader

from amulet_map_editor import log
from amulet.api.level import BaseLevel
from amulet.api.data_types import Dimension
from amulet.api.selection import SelectionGroup


class OperationLoader(BaseOperationLoader):
    def __init__(self, identifier: str, export_dict: dict):
        self._operation: Optional[
            Callable[[BaseLevel, Dimension, SelectionGroup], Any]
        ] = None
        super().__init__(identifier, export_dict)

    def _setup(self, export_dict: dict):
        """Parse the export dictionary and setup as required."""
        if "operation" in export_dict:
            if callable(export_dict["operation"]):
                operation = export_dict["operation"]
                if operation.__code__.co_argcount == 3:
                    self._operation = operation
                    self._is_valid = True
                else:
                    log.error(
                        f'"operation" function in export must have 3 inputs. {self.identifier}'
                    )
            else:
                log.error(
                    f'"operation" in export must be a callable. {self.identifier}'
                )
        else:
            log.error(f'"operation" is not present in export. {self.identifier}')

    def __call__(
        self, world: BaseLevel, dimension: Dimension, selection: SelectionGroup
    ):
        """Run the actual operation."""
        self._operation(world, dimension, selection)



from typing import Dict, Union, Callable, Tuple, Any

MenuData = Dict[
    str,
    Dict[
        str,
        Dict[
            str,
            Union[
                Callable,
                Tuple[Callable],
                Tuple[Callable, str],
                Tuple[Callable, str, Any],
            ],
        ],
    ],
]



from typing import TYPE_CHECKING
import wx

from amulet_map_editor.api.opengl.mesh.selection import RenderSelection
from amulet_map_editor.api.opengl.camera import Projection
from amulet.api.data_types import BlockCoordinatesNDArray, BlockCoordinates

from .raycast_behaviour import RaycastBehaviour
from ..events import (
    EVT_PRE_DRAW,
    EVT_CAMERA_MOVED,
    InputPressEvent,
    EVT_INPUT_PRESS,
)
from ..key_config import (
    ACT_INCR_SELECT_DISTANCE,
    ACT_DECR_SELECT_DISTANCE,
)

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas

_PointChangeEventType = wx.NewEventType()
EVT_POINT_CHANGE = wx.PyEventBinder(_PointChangeEventType)


class PointChangeEvent(wx.PyEvent):
    """Run when the cursor point changed."""

    def __init__(self, point: BlockCoordinates):
        wx.PyEvent.__init__(self, eventType=_PointChangeEventType)
        self._point = point

    @property
    def point(self) -> BlockCoordinates:
        return self._point


class PointerBehaviour(RaycastBehaviour):
    """Adds the behaviour of the selection pointer."""

    def __init__(self, canvas: "EditCanvas"):
        super().__init__(canvas)

        # has the pointer moved
        self._pointer_moved = True

        # the distance between the camera and the pointer
        self._pointer_distance = 10

        # the pointer
        self._pointer = RenderSelection(
            self.canvas.context_identifier,
            self.canvas.renderer.opengl_resource_pack,
        )

    @property
    def pointer_base(self) -> BlockCoordinatesNDArray:
        return self._pointer.point1

    def bind_events(self):
        super().bind_events()
        self.canvas.Bind(EVT_PRE_DRAW, self._pre_draw)
        self.canvas.Bind(EVT_CAMERA_MOVED, self._invalidate_pointer)
        self.canvas.Bind(wx.EVT_MOTION, self._invalidate_pointer)
        self.canvas.Bind(EVT_INPUT_PRESS, self._on_input_press)

    def _on_input_press(self, evt: InputPressEvent):
        if evt.action_id == ACT_INCR_SELECT_DISTANCE:
            self._pointer_distance += 1
            self._pointer_moved = True
        elif evt.action_id == ACT_DECR_SELECT_DISTANCE:
            self._pointer_distance -= 1
            self._pointer_moved = True
        evt.Skip()

    def _invalidate_pointer(self, evt):
        self._pointer_moved = True
        evt.Skip()

    def _pre_draw(self, evt):
        if self._pointer_moved:
            self._update_pointer()
            self._pointer_moved = False
        evt.Skip()

    def _post_change_event(self):
        wx.PostEvent(
            self.canvas, PointChangeEvent(tuple(self._pointer.point1.tolist()))
        )

    def _update_pointer(self):
        """Update the pointer location."""
        if self.canvas.camera.projection_mode == Projection.TOP_DOWN:
            location = self.closest_block_2d()[0]
        else:
            if self.canvas.camera.rotating:
                location = self.distance_block_3d(self._pointer_distance)
            else:
                location = self.closest_block_3d()[0]

        self._pointer.point1, self._pointer.point2 = location, location + 1
        self._post_change_event()

    def draw(self):
        self._pointer.draw(self.canvas.camera.transformation_matrix)



from typing import TYPE_CHECKING

from amulet.api.structure import structure_cache
from amulet.api.data_types import Dimension
from amulet.api.selection import SelectionGroup
from amulet_map_editor.programs.edit.api.operations import (
    OperationSilentAbort,
    OperationError,
)
from amulet.api.data_types import OperationReturnType

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel


def copy(
    world: "BaseLevel", dimension: Dimension, selection: SelectionGroup
) -> OperationReturnType:
    if selection:
        yield 0, "Copying"
        structure = yield from world.extract_structure_iter(selection, dimension)
        structure_cache.add_structure(structure, structure.dimensions[0])
        raise OperationSilentAbort
    else:
        raise OperationError(
            "At least one selection is required for the copy operation."
        )



import wx
import math

from amulet.api.data_types import FloatTriplet
from .select_location import SelectLocationUI
from .events import RotationChangeEvent, ScaleChangeEvent, TransformChangeEvent


class SelectTransformUI(SelectLocationUI):
    """A UI element that can be dropped into the EditCanvas and let the user pick the transform for a structure."""

    def _setup_ui(self):
        super()._setup_ui()
        # self._sx: wx.SpinCtrl = self._add_row('sx', wx.SpinCtrlDouble, min=-100, max=100, initial=1)
        # self._sy: wx.SpinCtrl = self._add_row('sy', wx.SpinCtrlDouble, min=-100, max=100, initial=1)
        # self._sz: wx.SpinCtrl = self._add_row('sz', wx.SpinCtrlDouble, min=-100, max=100, initial=1)
        # self._sx.Bind(wx.EVT_SPINCTRLDOUBLE, self._on_scale_change)
        # self._sy.Bind(wx.EVT_SPINCTRLDOUBLE, self._on_scale_change)
        # self._sz.Bind(wx.EVT_SPINCTRLDOUBLE, self._on_scale_change)

        self._rx: wx.SpinCtrl = self._add_row(
            "rx",
            wx.SpinCtrlDouble,
            min=-30000000,
            max=30000000,
            inc=90,
            style=wx.SP_ARROW_KEYS | wx.SP_WRAP,
        )
        self._ry: wx.SpinCtrl = self._add_row(
            "ry",
            wx.SpinCtrlDouble,
            min=-30000000,
            max=30000000,
            inc=90,
            style=wx.SP_ARROW_KEYS | wx.SP_WRAP,
        )
        self._rz: wx.SpinCtrl = self._add_row(
            "rz",
            wx.SpinCtrlDouble,
            min=-30000000,
            max=30000000,
            inc=90,
            style=wx.SP_ARROW_KEYS | wx.SP_WRAP,
        )
        self._rx.Bind(wx.EVT_SPINCTRLDOUBLE, self._on_rotation_change)
        self._ry.Bind(wx.EVT_SPINCTRLDOUBLE, self._on_rotation_change)
        self._rz.Bind(wx.EVT_SPINCTRLDOUBLE, self._on_rotation_change)

    @property
    def scale(self) -> FloatTriplet:
        return 1, 1, 1
        # return self._sx.GetValue(), self._sy.GetValue(), self._sz.GetValue()

    @property
    def rotation(self) -> FloatTriplet:
        return self._rx.GetValue(), self._ry.GetValue(), self._rz.GetValue()

    @property
    def rotation_radians(self) -> FloatTriplet:
        """The rotation of the object. (x, y, z)."""
        return tuple(math.radians(r) for r in self.rotation)

    def _on_rotation_change(self, evt):
        wx.PostEvent(self, RotationChangeEvent(self.rotation))
        self._on_transform_change()

    def _on_scale_change(self, evt):
        for ctrl in (self._rx, self._ry, self._rz):
            ctrl.SetValue(
                round((ctrl.GetValue() % 360) / 90) * 90
            )  # TODO: change this if smaller increments are allowed
        wx.PostEvent(self, ScaleChangeEvent(self.scale))
        self._on_transform_change()

    def _on_transform_change(self):
        wx.PostEvent(
            self, TransformChangeEvent(self.location, self.rotation, self.scale)
        )



from typing import TYPE_CHECKING, Tuple
import wx
import numpy

from amulet.api.block import Block

from amulet_map_editor.api.wx.ui.base_select import EVT_PICK
from amulet_map_editor.api.wx.ui.block_select import BlockDefine
from amulet_map_editor.api.wx.ui.simple import SimpleScrollablePanel
from amulet_map_editor.programs.edit.api.operations import DefaultOperationUI

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class Replace(SimpleScrollablePanel, DefaultOperationUI):
    def __init__(
        self,
        parent: wx.Window,
        canvas: "EditCanvas",
        world: "BaseLevel",
        options_path: str,
    ):
        SimpleScrollablePanel.__init__(self, parent)
        DefaultOperationUI.__init__(self, parent, canvas, world, options_path)
        self.Freeze()
        options = self._load_options({})

        self._original_block = BlockDefine(
            self,
            world.level_wrapper.translation_manager,
            wx.VERTICAL,
            *(
                options.get("original_block_options", [])
                or [world.level_wrapper.platform]
            ),
            wildcard_properties=True,
            show_pick_block=True
        )
        self._sizer.Add(self._original_block, 1, wx.ALL | wx.ALIGN_CENTRE_HORIZONTAL, 5)
        self._original_block.Bind(EVT_PICK, lambda evt: self._on_pick_block_button(1))
        self._replacement_block = BlockDefine(
            self,
            world.level_wrapper.translation_manager,
            wx.VERTICAL,
            *(
                options.get("replacement_block_options", [])
                or [world.level_wrapper.platform]
            ),
            show_pick_block=True
        )
        self._sizer.Add(
            self._replacement_block, 1, wx.ALL | wx.ALIGN_CENTRE_HORIZONTAL, 5
        )
        self._replacement_block.Bind(
            EVT_PICK, lambda evt: self._on_pick_block_button(2)
        )

        self._run_button = wx.Button(self, label="Run Operation")
        self._run_button.Bind(wx.EVT_BUTTON, self._run_operation)
        self._sizer.Add(self._run_button, 0, wx.ALL | wx.ALIGN_CENTRE_HORIZONTAL, 5)

        self.Layout()
        self.Thaw()

    @property
    def wx_add_options(self) -> Tuple[int, ...]:
        return (1,)

    def _on_pick_block_button(self, code):
        """Set up listening for the block click"""
        self._show_pointer = code

    def _on_box_click(self):
        if self._show_pointer:
            x, y, z = self._pointer.pointer_base
            if self._show_pointer == 1:
                self._original_block.universal_block = (
                    self.world.get_block(x, y, z, self.canvas.dimension),
                    None,
                )
            elif self._show_pointer == 2:
                self._replacement_block.universal_block = (
                    self.world.get_block(x, y, z, self.canvas.dimension),
                    None,
                )
            self._show_pointer = False

    def _get_replacement_block(self) -> Block:
        return self._replacement_block.universal_block[0]

    def disable(self):
        self._save_options(
            {
                "original_block_options": (
                    self._original_block.platform,
                    self._original_block.version_number,
                    self._original_block.force_blockstate,
                    self._original_block.namespace,
                    self._original_block.block_name,
                    self._original_block.str_properties,
                ),
                "replacement_block": self._get_replacement_block(),
                "replacement_block_options": (
                    self._replacement_block.platform,
                    self._replacement_block.version_number,
                    self._replacement_block.force_blockstate,
                    self._replacement_block.namespace,
                    self._replacement_block.block_name,
                    self._replacement_block.str_properties,
                ),
            }
        )

    def _run_operation(self, _):
        self.canvas.run_operation(lambda: self._replace())

    def _replace(self):
        world = self.world
        selection = self.canvas.selection.selection_group
        dimension = self.canvas.dimension

        (
            original_platform,
            original_version,
            original_blockstate,
            original_namespace,
            original_base_name,
            original_properties,
        ) = (
            self._original_block.platform,
            self._original_block.version_number,
            self._original_block.force_blockstate,
            self._original_block.namespace,
            self._original_block.block_name,
            self._original_block.str_properties,
        )
        replacement_block = self._get_replacement_block()

        replacement_block_id = world.block_palette.get_add_block(replacement_block)

        original_block_matches = []
        universal_block_count = 0

        iter_count = len(list(world.get_coord_box(dimension, selection)))
        count = 0

        for chunk, slices, _ in world.get_chunk_slice_box(dimension, selection):
            if universal_block_count < len(world.block_palette):
                for universal_block_id in range(
                    universal_block_count, len(world.block_palette)
                ):
                    version_block = world.translation_manager.get_version(
                        original_platform, original_version
                    ).block.from_universal(
                        world.block_palette[universal_block_id],
                        force_blockstate=original_blockstate,
                    )[
                        0
                    ]
                    if (
                        version_block.namespace == original_namespace
                        and version_block.base_name == original_base_name
                        and all(
                            original_properties.get(prop) in ["*", val.to_snbt()]
                            for prop, val in version_block.properties.items()
                        )
                    ):
                        original_block_matches.append(universal_block_id)

                universal_block_count = len(world.block_palette)
            blocks = chunk.blocks[slices]
            blocks[numpy.isin(blocks, original_block_matches)] = replacement_block_id
            chunk.blocks[slices] = blocks
            chunk.changed = True

            count += 1
            yield count / iter_count

    def DoGetBestClientSize(self):
        sizer = self.GetSizer()
        if sizer is None:
            return -1, -1
        else:
            sx, sy = self.GetSizer().CalcMin()
            return (
                sx + wx.SystemSettings.GetMetric(wx.SYS_VSCROLL_X),
                sy + wx.SystemSettings.GetMetric(wx.SYS_HSCROLL_Y),
            )


export = {
    "name": "Replace",  # the name of the plugin
    "operation": Replace,  # the actual function to call when running the plugin
}



from __future__ import annotations

import urllib.request
import threading
import json
import re
import webbrowser
from dataclasses import dataclass
from enum import IntEnum

import wx

from amulet_map_editor import lang

URL = "http://api.github.com/repos/Amulet-Team/Amulet-Map-Editor/releases"

"""
VERSION_REGEX Documentation

    This is not a documentation of regular expression syntax, just documentation on the subsequences that this regular expression looks for
    For an in-depth guide on regular expressions, see here: https://www.debuggex.com/cheatsheet/regex/python
    
    Our version scheme allows for an optional "v" prefix, so that subsequence is tested by "v?"
    
    Since we use the Semantic Versioning scheme, our version numbers follow this format: <major>.<minor>.<patch>
    with the <patch> number being optional (only if it's value would be 0 for that version)
    That is tested by "(?P<major>\d+)\.(?P<minor>\d+)(\.(?P<patch>\d+))?"
    
    - For the major and minor numbers, "(?P<major>\d+)" will match any non-zero number of digits and group them under the
        "major" group (in this instance)
    - "\." will then match the separating "." character
    - Since the patch number is optional, "(\.(?P<patch>\d+))?" is used instead, the second "." is included in the 
        optional group, but other wise the capturing group is the same as the major/minor numbers
        
    We then use an optional release type flag to denote whether the release is a beta or alpha release and it's number:
    "((a(?P<alpha>\d+))|(b(?P<beta>\d+)))?"
    
    For our nightly builds, we also denote a unique build number with the prefix of ".dev": "(\.dev(?P<devnum>\d{12}))?"
    
    Finally, our version scheme also denotes when Amulet is being ran from source, while this information isn't
    necessarily used for update checking, we use it to determine if the update dialog should be shown. If a commit 
    hash/count is found in the version string of the current Amulet instance, the dialog is not shown. Our regular 
    expression uses the following to detect that subsequence: 
        "(\+(?P<commit_count>\d+)\.g(?P<commit_hash>[a-z\d]+)(\.dirty)?)?"
        
    - "commit_count" group signifies the number of commits since the last release
    - "commit_hash" group is the current commit hash
    - ".dirty" denotes whether uncommitted changes have been made 
"""
VERSION_REGEX = re.compile(
    r"^v?(?P<major>\d+)\.(?P<minor>\d+)(\.(?P<patch>\d+))?((a(?P<alpha>\d+))|(b(?P<beta>\d+)))?(\.dev(?P<devnum>\d{12}))?(\+(?P<commit_count>\d+)\.g(?P<commit_hash>[a-z\d]+)(\.dirty)?)?$"
)

_EVT_UPDATE_CHECK = wx.NewEventType()
EVT_UPDATE_CHECK = wx.PyEventBinder(_EVT_UPDATE_CHECK, 1)


class Release(IntEnum):
    FULL = 0
    BETA = -1
    ALPHA = -2


@dataclass
class Version:
    release_stage: int
    major: int
    minor: int
    patch: int
    alpha_number: int = -1
    beta_number: int = -1
    nightly_timestamp: int = -1
    has_commit_hash: bool = False

    def __gt__(self, other: Version):
        if self.version_tuple > other.version_tuple:
            return True
        elif self.version_tuple == other.version_tuple:
            if self.release_stage > other.release_stage:
                return True
            elif self.release_stage == other.release_stage:
                if self.release_stage == Release.ALPHA:
                    return self.alpha_number > other.alpha_number
                elif self.release_stage == Release.BETA:
                    if self.beta_number > other.beta_number:
                        return True
                    elif (
                        self.beta_number == other.beta_number and self.beta_number != -1
                    ):
                        return self.nightly_timestamp > other.nightly_timestamp
        return False

    @property
    def version_tuple(self):
        return self.major, self.minor, self.patch


def get_version(version_string: str) -> Version:
    """Parse the version into a more usable format

    :param version_string: The version string. Eg 1.2 or 1.2.3.4 or 1.2.3.4b0
    :return: A Version object from the parsed version string
    """
    version_match = VERSION_REGEX.match(version_string)
    if version_match:
        v = version_match.groupdict()
        major, minor = int(v["major"]), int(v["minor"])
        if v["patch"] is None:
            patch = 0
        else:
            patch = int(v["patch"])
        version = Version(Release.FULL, major, minor, patch)

        if v.get("alpha") is not None:
            version.release_stage = Release.ALPHA
            version.alpha_number = int(v["alpha"])
        elif v.get("beta") is not None:
            version.release_stage = Release.BETA
            version.beta_number = int(v["beta"])
            if v.get("devnum") is not None:
                version.nightly_timestamp = int(v["devnum"])

        if v.get("commit_hash") is not None:
            version.has_commit_hash = True
        return version

    raise Exception(f"Invalid version string {version_string}")


class UpdateEvent(wx.PyCommandEvent):
    def __init__(self, etype, eid, new_version: str):
        wx.PyCommandEvent.__init__(self, etype, eid)
        self._new_version = new_version

    def GetVersion(self) -> str:
        return self._new_version


class CheckForUpdate(threading.Thread):
    def __init__(self, url, current_version, parent):
        threading.Thread.__init__(self)
        self.url = url
        self.current_version = current_version
        self._parent = parent
        self._new_version = None

    def run(self):
        try:
            conn = urllib.request.urlopen(self.url, timeout=5)
            data = conn.read()
            data = json.loads(data)
            try:
                current_version = get_version(self.current_version)
            except Exception:
                return

            for release_version, release_data in sorted(
                map(lambda d: (get_version(d["tag_name"]), d), data),
                key=lambda t: t[0],
                reverse=True,
            ):
                if (
                    release_version > current_version
                    and release_version.release_stage >= current_version.release_stage
                ):
                    self._new_version = release_data["tag_name"]
                    break

            if self._new_version:
                evt = UpdateEvent(_EVT_UPDATE_CHECK, -1, self._new_version)
                wx.PostEvent(self._parent, evt)
        except Exception:
            pass


class UpdateDialog(wx.Dialog):
    def __init__(self, parent, current_version: str, new_version: str):
        wx.Dialog.__init__(self, parent)

        sizer_1 = wx.BoxSizer(wx.VERTICAL)

        static_text_1 = wx.StaticText(
            self, label=lang.get("update_check.newer_version_released")
        )
        static_text_2 = wx.StaticText(
            self, label=f"{lang.get('update_check.new_version')} v{new_version}"
        )
        static_text_3 = wx.StaticText(
            self, label=f"{lang.get('update_check.current_version')} v{current_version}"
        )

        sizer_1.Add(static_text_1, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)
        sizer_1.Add(static_text_2, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)
        sizer_1.Add(static_text_3, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)

        sizer_2 = wx.BoxSizer(wx.HORIZONTAL)

        update_button = wx.Button(self, label=lang.get("update_check.update"))
        ok_button = wx.Button(self, label=lang.get("update_check.ok"))

        sizer_2.Add(update_button, 0, wx.ALL, 5)
        sizer_2.Add((0, 0), 1, wx.EXPAND, 5)
        sizer_2.Add(ok_button, 0, wx.ALL, 5)

        sizer_1.Add(sizer_2, 1, wx.EXPAND, 5)

        self.SetSizerAndFit(sizer_1)

        self.Centre()

        update_button.Bind(
            wx.EVT_BUTTON, lambda evt: self.goto_download_page(new_version, evt)
        )
        ok_button.Bind(wx.EVT_BUTTON, lambda evt: self.Close())

    @staticmethod
    def goto_download_page(new_version, _):
        webbrowser.open(
            f"https://github.com/Amulet-Team/Amulet-Map-Editor/releases/tag/{new_version}"
        )


def show_update_window(parent, current_version: str, evt: UpdateEvent):
    if get_version(current_version).has_commit_hash:
        print("Running from source, not showing update dialog")
        return
    UpdateDialog(parent, current_version, evt.GetVersion()).ShowModal()


def check_for_update(version, listening_parent):
    update_thread = CheckForUpdate(URL, version, listening_parent)
    update_thread.start()



import logging
import sys
import os

log = logging.getLogger("amulet_map_editor")
log_level = logging.DEBUG if "amulet-debug" in sys.argv else logging.INFO

log.setLevel(log_level)

_formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")

os.makedirs("./logs", exist_ok=True)
_log_file = logging.FileHandler("./logs/amulet_map_editor.log", "w", encoding="utf-8")
_log_file.setLevel(log_level)
_log_file.setFormatter(_formatter)
log.addHandler(_log_file)

_log_console = logging.StreamHandler()
_log_console.setLevel(log_level)
_log_console.setFormatter(_formatter)
log.addHandler(_log_console)



from .base import OperationUI, OperationUIType, DefaultOperationUI
from .ui.fixed_pipeline import FixedFunctionUI
from .errors import OperationError, OperationSuccessful, OperationSilentAbort
from .ui.simple_operation_panel import SimpleOperationPanel



import numpy
from typing import TYPE_CHECKING, Tuple, List, Union
import weakref
import itertools
from amulet_map_editor import log

from amulet.api.errors import ChunkLoadError, ChunkDoesNotExist
from amulet.api.chunk.blocks import Blocks
from amulet.api.data_types import Dimension
from amulet.api.level import BaseLevel
from amulet.api.selection import SelectionBox

from .chunk_builder import RenderChunkBuilder
from amulet_map_editor.api.opengl.resource_pack import OpenGLResourcePack

if TYPE_CHECKING:
    from amulet.api.chunk import Chunk


class RenderChunk(RenderChunkBuilder):
    def __init__(
        self,
        context_identifier: str,
        resource_pack: OpenGLResourcePack,
        level: BaseLevel,
        region_size: int,
        chunk_coords: Tuple[int, int],
        dimension: Dimension,
        draw_floor: bool = False,
        draw_ceil: bool = False,
        limit_bounds: bool = False,
    ):
        # the chunk geometry is stored in chunk space (floating point)
        # at shader time it is transformed by the players transform
        super().__init__(context_identifier, resource_pack)
        self._level_ = weakref.ref(level)
        self._region_size = region_size
        self._coords = chunk_coords
        self._dimension = dimension
        self._draw_floor = draw_floor
        self._draw_ceil = draw_ceil
        self._limit_bounds = limit_bounds
        self._chunk_state = 0  # 0 = chunk does not exist, 1 = chunk exists but failed to load, 2 = chunk exists
        self._changed_time = 0
        self._needs_rebuild = True
        self.verts_translucent = (
            0  # the offset into the above from which the faces can be translucent
        )
        # self.chunk_lod1: numpy.ndarray = self.new_empty_verts()

    def __repr__(self):
        return f"RenderChunk({self._coords[0]}, {self._coords[1]})"

    def _setup(self):
        """Set up the opengl data which cannot be set up in another thread"""
        super()._setup()
        if self._needs_rebuild:
            self.change_verts()
            self._needs_rebuild = False

    @property
    def _level(self) -> BaseLevel:
        return self._level_()

    @property
    def offset(self) -> numpy.ndarray:
        return 16 * (
            numpy.array([self._coords[0], 0, self._coords[1]]) % self._region_size
        )

    @property
    def dimension(self) -> str:
        return self._dimension

    @property
    def cx(self) -> int:
        return self._coords[0]

    @property
    def cz(self) -> int:
        return self._coords[1]

    @property
    def coords(self) -> Tuple[int, int]:
        return self._coords

    @property
    def chunk(self) -> "Chunk":
        return self._level.get_chunk(self.cx, self.cz, self._dimension)

    @property
    def chunk_state(self) -> int:
        return self._chunk_state

    def needs_rebuild(self):
        """has the chunk data changed since the last rebuild"""
        try:
            chunk = self.chunk
        except ChunkDoesNotExist:
            chunk_state = 0
        except ChunkLoadError:
            chunk_state = 1
        else:
            chunk_state = 2
            if chunk.changed_time != self._changed_time:
                return True
        return chunk_state != self._chunk_state

    def _sub_chunks(
        self, blocks: Blocks
    ) -> List[Tuple[numpy.ndarray, Tuple[int, int, int]]]:
        """Create sub-chunk arrays that extend into the neighbour sub-chunks by one block.

        :param blocks: The Blocks array for the chunk.
        :return: A list of tuples containing the larger block array and the location of the sub-chunk
        """
        sub_chunks = []
        neighbour_chunks = {}
        for dx, dz in ((-1, 0), (1, 0), (0, -1), (0, 1)):
            try:
                neighbour_chunks[(dx, dz)] = self._level.get_chunk(
                    self.cx + dx, self.cz + dz, self.dimension
                ).blocks
            except ChunkLoadError:
                continue

        for cy in blocks.sub_chunks:
            sub_chunk = blocks.get_sub_chunk(cy)
            larger_blocks = numpy.zeros(
                sub_chunk.shape + numpy.array((2, 2, 2)), sub_chunk.dtype
            )
            sub_chunk_box = SelectionBox.create_sub_chunk_box(self.cx, cy, self.cz)
            if self._limit_bounds:
                if self._level.bounds(self.dimension).intersects(sub_chunk_box):
                    boxes = self._level.bounds(self.dimension).intersection(
                        sub_chunk_box
                    )
                    for box in boxes.selection_boxes:
                        larger_blocks[1:-1, 1:-1, 1:-1][
                            box.sub_chunk_slice(self.cx, cy, self.cz)
                        ] = sub_chunk[box.sub_chunk_slice(self.cx, cy, self.cz)]
                else:
                    continue
            else:
                larger_blocks[1:-1, 1:-1, 1:-1] = sub_chunk
            for chunk_offset, neighbour_blocks in neighbour_chunks.items():
                if cy not in neighbour_blocks:
                    continue
                if chunk_offset == (-1, 0):
                    larger_blocks[0, 1:-1, 1:-1] = neighbour_blocks.get_sub_chunk(cy)[
                        -1, :, :
                    ]
                elif chunk_offset == (1, 0):
                    larger_blocks[-1, 1:-1, 1:-1] = neighbour_blocks.get_sub_chunk(cy)[
                        0, :, :
                    ]
                elif chunk_offset == (0, -1):
                    larger_blocks[1:-1, 1:-1, 0] = neighbour_blocks.get_sub_chunk(cy)[
                        :, :, -1
                    ]
                elif chunk_offset == (0, 1):
                    larger_blocks[1:-1, 1:-1, -1] = neighbour_blocks.get_sub_chunk(cy)[
                        :, :, 0
                    ]
            if cy - 1 in blocks:
                larger_blocks[1:-1, 0, 1:-1] = blocks.get_sub_chunk(cy - 1)[:, -1, :]
            if cy + 1 in blocks:
                larger_blocks[1:-1, -1, 1:-1] = blocks.get_sub_chunk(cy + 1)[:, 0, :]
            sub_chunks.append((larger_blocks, (0, cy * 16, 0)))
        return sub_chunks

    def create_geometry(self):
        try:
            chunk = self.chunk
        except ChunkDoesNotExist:
            self._create_empty_geometry()
            self._chunk_state = 0
        except ChunkLoadError:
            log.info(f"Error loading chunk {self.coords}", exc_info=True)
            self._create_error_geometry()
            self._chunk_state = 1
        else:
            self._changed_time = chunk.changed_time
            self._chunk_state = 2
            chunk_verts, chunk_verts_translucent = self._create_lod0_multi(
                self._sub_chunks(chunk.blocks)
            )
            self._set_verts(chunk_verts, chunk_verts_translucent)
            if self._draw_floor or self._draw_ceil:
                plane = self._create_grid(
                    "amulet",
                    "amulet_ui/translucent_white",
                    (0.55, 0.5, 0.9) if (self.cx + self.cz) % 2 else (0.4, 0.4, 0.85),
                )
                self.verts = numpy.concatenate([self.verts, plane.ravel()], 0)
                self.draw_count += len(plane)
        self._needs_rebuild = True

    def _create_empty_geometry(self):
        if self._draw_floor:
            plane = self._create_grid(
                "amulet",
                "amulet_ui/translucent_white",
                (0.3, 0.3, 0.3) if (self.cx + self.cz) % 2 else (0.2, 0.2, 0.2),
            )
            self.verts = plane.ravel()
            self.draw_count = len(plane)
        else:
            self.verts = numpy.ones(0, numpy.float32)
            self.draw_count = 0

    def _create_grid(
        self,
        texture_namespace: str,
        texture_path: str,
        tint: Tuple[float, float, float],
    ):
        plane: numpy.ndarray = numpy.ones(
            (self._vert_len * 12 * (self._draw_floor + self._draw_ceil)),
            dtype=numpy.float32,
        ).reshape((-1, self._vert_len))
        bounds = self._level.bounds(self.dimension)
        if self._draw_floor:
            plane[:12, :3], plane[:12, 3:5] = self._create_chunk_plane(
                bounds.min_y - 0.01
            )
            if self._draw_ceil:
                plane[12:, :3], plane[12:, 3:5] = self._create_chunk_plane(
                    bounds.max_y + 0.01
                )
        elif self._draw_ceil:
            plane[:12, :3], plane[:12, 3:5] = self._create_chunk_plane(
                bounds.max_y + 0.01
            )

        plane[:, 5:9] = self.resource_pack.texture_bounds(
            self.resource_pack.get_texture_path(texture_namespace, texture_path)
        )
        plane[:, 9:12] = tint
        return plane

    def _create_chunk_plane(
        self, height: Union[int, float]
    ) -> Tuple[numpy.ndarray, numpy.ndarray]:
        box = numpy.array([(0, height, 0), (16, height, 16)]) + self.offset
        _box_coordinates = numpy.array(list(itertools.product(*box.T.tolist())))
        _cube_face_lut = numpy.array(
            [  # This maps to the verticies used (defined in cube_vert_lut)
                0,
                4,
                5,
                1,
                3,
                7,
                6,
                2,
            ]
        )
        box = box.ravel()
        _texture_index = numpy.array([0, 2, 3, 5, 0, 2, 3, 5], numpy.uint32)
        _uv_slice = numpy.array(
            [0, 1, 2, 1, 2, 3, 0, 3] * 2, dtype=numpy.uint32
        ).reshape((-1, 8)) + numpy.arange(0, 8, 4).reshape((-1, 1))

        _tri_face = numpy.array([0, 1, 2, 0, 2, 3] * 2, numpy.uint32).reshape(
            (-1, 6)
        ) + numpy.arange(0, 8, 4).reshape((-1, 1))
        return (
            _box_coordinates[_cube_face_lut[_tri_face]].reshape((-1, 3)),
            box[_texture_index[_uv_slice]]
            .reshape(-1, 2)[_tri_face, :]
            .reshape((-1, 2)),
        )

    def _create_error_geometry(self):
        if self._draw_floor:
            plane = self._create_grid(
                "amulet",
                "amulet_ui/translucent_white",
                (1, 0.2, 0.2) if (self.cx + self.cz) % 2 else (0.75, 0.2, 0.2),
            )
            self.verts = plane.ravel()
            self.draw_count = len(plane)
        else:
            self.verts = numpy.ones(0, numpy.float32)
            self.draw_count = 0

    def _create_lod1(
        self,
        blocks: numpy.ndarray,
        larger_blocks: numpy.ndarray,
        unique_blocks: numpy.ndarray,
    ):
        # TODO
        self.verts: numpy.ndarray = self.new_empty_verts()
        # self.chunk_lod1: numpy.ndarray = self.new_empty_verts()



from typing import TYPE_CHECKING

import wx
import math

from amulet.utils import block_coords_to_chunk_coords
from amulet.api.chunk.biomes import BiomesShape
from amulet_map_editor.api.wx.ui.base_select import EVT_PICK
from amulet_map_editor.api.wx.ui.biome_select import BiomeDefine
from amulet_map_editor.programs.edit.api.operations import SimpleOperationPanel
from amulet_map_editor.api.wx.ui.simple import SimpleChoiceAny

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas
    from amulet.api.selection import SelectionGroup
    from amulet.api.data_types import Dimension, OperationReturnType


BoxMode = "box"
ColumnMode = "column"

lang = {
    ColumnMode: "Selected Column",
    BoxMode: "Selected Box",
}


MODES = {
    ColumnMode: "Set the biomes for the vertical column the selection intersects.",
    BoxMode: "Set the biomes for only the blocks in the selection.",
}

Border = wx.TOP | wx.LEFT | wx.RIGHT | wx.EXPAND


class SetBiome(SimpleOperationPanel):
    def __init__(
        self,
        parent: wx.Window,
        canvas: "EditCanvas",
        world: "BaseLevel",
        options_path: str,
    ):
        SimpleOperationPanel.__init__(self, parent, canvas, world, options_path)
        self.Freeze()
        options = self._load_options({})

        self._mode = SimpleChoiceAny(self, sort=False)
        self._mode.SetItems({mode: lang[mode] for mode in MODES.keys()})
        self._sizer.Add(self._mode, 0, Border, 5)
        self._mode.Bind(wx.EVT_CHOICE, self._on_mode_change)

        self._mode_description = wx.TextCtrl(
            self, style=wx.TE_MULTILINE | wx.TE_READONLY | wx.TE_BESTWRAP
        )
        self._sizer.Add(self._mode_description, 0, Border, 5)

        self._mode_description.SetLabel(MODES[self._mode.GetCurrentObject()])
        self._mode_description.Fit()

        self._biome_choice = BiomeDefine(
            self,
            world.level_wrapper.translation_manager,
            wx.VERTICAL,
            *(
                options.get("original_block_options", [])
                or [world.level_wrapper.platform]
            ),
            show_pick_biome=True,
        )
        self._biome_choice.Bind(EVT_PICK, self._on_pick_biome_button)
        self._sizer.Add(self._biome_choice, 1, Border, 5)

        self._add_run_button()

        self.Thaw()

    def _on_mode_change(self, evt):
        self._mode_description.SetLabel(MODES[self._mode.GetCurrentObject()])
        self._mode_description.Fit()
        self.Layout()
        evt.Skip()

    def _on_pick_biome_button(self, evt):
        """Set up listening for the biome click"""
        self._show_pointer = True

    def _on_box_click(self):
        if self._show_pointer:
            self._show_pointer = False
            x, y, z = self._pointer.pointer_base

            # TODO: replace with "get_biome(x, y, z)" if it'll be created
            cx, cz = block_coords_to_chunk_coords(
                x, z, sub_chunk_size=self.world.sub_chunk_size
            )
            offset_x, offset_z = x - 16 * cx, z - 16 * cz
            chunk = self.world.get_chunk(cx, cz, self.canvas.dimension)

            if chunk.biomes.dimension == BiomesShape.Shape3D:
                biome = chunk.biomes[offset_x // 4, y // 4, offset_z // 4]
            elif chunk.biomes.dimension == BiomesShape.Shape2D:
                biome = chunk.biomes[offset_x, offset_z]
            else:
                return

            self._biome_choice.universal_biome = chunk.biome_palette[biome]

    def _operation(
        self, world: "BaseLevel", dimension: "Dimension", selection: "SelectionGroup"
    ) -> "OperationReturnType":
        mode = self._mode.GetCurrentObject()

        iter_count = len(list(world.get_coord_box(dimension, selection, False)))
        for count, (chunk, slices, _) in enumerate(
            world.get_chunk_slice_box(dimension, selection, False)
        ):
            new_biome = chunk.biome_palette.get_add_biome(
                self._biome_choice.universal_biome
            )

            if mode == BoxMode:
                if chunk.biomes.dimension == BiomesShape.Shape3D:
                    slices = (
                        slice(slices[0].start // 4, math.ceil(slices[0].stop / 4)),
                        slice(slices[1].start // 4, math.ceil(slices[1].stop / 4)),
                        slice(slices[2].start // 4, math.ceil(slices[2].stop / 4)),
                    )
                elif chunk.biomes.dimension == BiomesShape.Shape2D:
                    slices = (slices[0], slices[2])
                else:
                    continue
            elif mode == ColumnMode:
                if chunk.biomes.dimension == BiomesShape.Shape3D:
                    slices = (
                        slice(slices[0].start // 4, math.ceil(slices[0].stop / 4)),
                        slice(None, None, None),
                        slice(slices[2].start // 4, math.ceil(slices[2].stop / 4)),
                    )
                elif chunk.biomes.dimension == BiomesShape.Shape2D:
                    slices = (slices[0], slices[2])
                else:
                    continue
            else:
                raise ValueError(
                    f"mode {mode} is not a valid mode for the Set Biome operation."
                )

            chunk.biomes[slices] = new_biome

            chunk.changed = True
            yield (count + 1) / iter_count


export = {
    "name": "Set Biome",  # the name of the plugin
    "operation": SetBiome,  # the actual function to call when running the plugin
}



import wx
from typing import Tuple
from .window_container import WindowContainer


class MouseMovement(WindowContainer):
    """A class to get and set the cursor position and track changes."""

    def __init__(self, window: wx.Window):
        super().__init__(window)
        # the current mouse position
        self._x: float = 0
        self._y: float = 0

        # the resting position of the mouse from which to calculate the delta
        # if the mouse is warped this will become the warped point
        self._start_x: float = 0
        self._start_y: float = 0

        # the sum of all the deltas since it was last reset.
        # when the mouse is warped the delta before it was warped will be added to this value
        self._delta_x: float = 0
        self._delta_y: float = 0

    def bind_events(self):
        """Set up all events required to run."""
        self.window.Bind(wx.EVT_MOTION, self._on_mouse_motion)

    def _to_relative(self, x: int, y: int) -> Tuple[float, float]:
        """Convert the x and y values to relative values 0 to 1"""
        dx, dy = self.window.GetSize()
        dx = max(1, dx)
        dy = max(1, dy)
        return x / dx, y / dy

    def _to_absolute(self, x: float, y: float) -> Tuple[int, int]:
        """Convert the relative values to absolute values."""
        dx, dy = self.window.GetSize()
        return int(x * dx), int(y * dy)

    def _on_mouse_motion(self, evt):
        """Event fired when the mouse is moved."""
        self._x, self._y = self._to_relative(*evt.GetPosition())
        evt.Skip()

    def _screen_middle(self) -> Tuple[int, int]:
        """Get the pixel coordinate of the middle of the screen"""
        x, y = self.window.GetSize()
        return (int(x / 2), int(y / 2))

    def warp_middle(self):
        """Warp the cursor to the middle of the screen."""
        self.xy = self._screen_middle()

    def set_middle(self):
        """Set the start point to the middle of the screen. Does not warp."""
        self._start_x, self._start_y = 0.5, 0.5

    def _warp(self):
        self.window.WarpPointer(*self._to_absolute(self._x, self._y))

    @property
    def x(self) -> int:
        """The x pixel location of the mouse in the parent window."""
        return self._to_absolute(self._x, 0)[0]

    @x.setter
    def x(self, x: int):
        """Set the x pixel location of the mouse in the parent window.
        Will warp the cursor to this position and create a mouse move event."""
        self._delta_x += self._x - self._start_x
        self._start_x = self._x = self._to_relative(int(x), 0)[0]
        self._warp()

    @property
    def y(self) -> int:
        """The y pixel location of the mouse in the parent window."""
        return self._to_absolute(0, self._y)[1]

    @y.setter
    def y(self, y: int):
        """Set the y pixel location of the mouse in the parent window.
        Will warp the cursor to this position and create a mouse move event."""
        self._delta_y += self._y - self._start_y
        self._start_y = self._y = self._to_relative(0, int(y))[1]
        self._warp()

    @property
    def xy(self) -> Tuple[int, int]:
        """The x and y pixel location of the mouse in the parent window."""
        return self._to_absolute(self._x, self._y)

    @xy.setter
    def xy(self, xy: Tuple[int, int]):
        """Set the x and y pixel location of the mouse in the parent window.
        Will warp the cursor to this position and create a mouse move event."""
        assert len(xy) == 2, "xy must be an iterable of two ints"
        x, y = map(int, xy)
        self._delta_x += self._x - self._start_x
        self._delta_y += self._y - self._start_y
        (self._start_x, self._start_y) = (self._x, self._y) = self._to_relative(x, y)
        self._warp()

    def reset_delta(self):
        """Reset the mouse delta values.
        The amount the mouse has moved will be reset and the start position will become the current position.
        :return:
        """
        self._delta_x = self._delta_y = 0
        self._start_x = self._x
        self._start_y = self._y

    @property
    def delta_x(self) -> int:
        """The x pixel distance between the current location and the location when reset_delta was called.
        If the pointer was warped the offset before it was warped will be added to this."""
        return self.delta_xy[0]

    @property
    def delta_y(self) -> int:
        """The y pixel distance between the current location and the location when reset_delta was called.
        If the pointer was warped the offset before it was warped will be added to this."""
        return self.delta_xy[1]

    @property
    def delta_xy(self) -> Tuple[int, int]:
        """The x and y pixel distance between the current location and the location when reset_delta was called.
        If the pointer was warped the offset before it was warped will be added to this."""
        return self._to_absolute(
            self._x - self._start_x + self._delta_x,
            self._y - self._start_y + self._delta_y,
        )



from amulet_map_editor.api.wx.ui.base_select import BaseSelect


class BiomeSelect(BaseSelect):
    TypeName = "Biome"

    def _populate_namespace(self):
        version = self._translation_manager.get_version(
            self._platform, self._version_number
        )
        namespaces = list(
            set(
                [biome_id[: biome_id.find(":")] for biome_id in version.biome.biome_ids]
            )
        )
        self._do_text_event = False
        self._namespace_combo.Set(namespaces)

    def _populate_item_name(self):
        version = self._translation_manager.get_version(
            self._platform, self._version_number
        )
        self._names = [
            biome_id[len(self.namespace) + 1 :]
            for biome_id in version.biome.biome_ids
            if biome_id.startswith(self.namespace)
        ]
        self._list_box.SetItems(self._names)



from typing import TYPE_CHECKING, Optional, Tuple, Generator, List
import numpy
import math

from amulet.api.chunk import Chunk
from amulet.api.block import UniversalAirLikeBlocks
from amulet.api.errors import ChunkLoadError

from amulet_map_editor.api.opengl.matrix import rotation_matrix_xy

from amulet.api.data_types import PointCoordinatesNDArray

from .base_behaviour import BaseBehaviour

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class RaycastBehaviour(BaseBehaviour):
    """Adds the base behaviour for behaviours that needs to do ray casting."""

    def __init__(self, canvas: "EditCanvas"):
        super().__init__(canvas)

    def look_vector(self) -> numpy.ndarray:
        """
        The x,y,z vector for the direction the camera is facing
        :return: (x, y, z) numpy float array ranging from -1 to 1
        """
        look_vector = numpy.array([0, 0, 1, 0])

        if self.canvas.mouse.delta_xy != (0, 0):
            screen_width, screen_height = numpy.array(self.canvas.GetSize(), int) / 2
            screen_width = max(1, screen_width)
            screen_height = max(1, screen_height)
            screen_dx = math.atan(
                self.canvas.mouse.delta_x
                * self.canvas.camera.aspect_ratio
                * math.tan(math.radians(self.canvas.camera.fov / 2))
                / screen_width
            )
            screen_dy = math.atan(
                self.canvas.mouse.delta_y
                * math.cos(screen_dx)
                * math.tan(math.radians(self.canvas.camera.fov / 2))
                / screen_height
            )
            look_vector = numpy.matmul(
                rotation_matrix_xy(screen_dy, -screen_dx),
                look_vector,
            )
        ry, rx = self.canvas.camera.rotation
        look_vector = numpy.matmul(
            rotation_matrix_xy(*numpy.radians([rx, -ry])), look_vector
        )[:3]
        look_vector[abs(look_vector) < 0.000001] = 0.000001
        return look_vector

    def closest_block_3d(
        self, max_distance: float = 100
    ) -> Tuple[PointCoordinatesNDArray, bool]:
        """Find the location of the closest non-air block.

        If the end is reached an a non-air block was not found the end location will be returned.
        :param max_distance: The distance to search up to.
        :return: Tuple[The block coordinate, was a non-air block found in the range]
        """
        cx: Optional[int] = None
        cz: Optional[int] = None
        chunk: Optional[Chunk] = None
        in_air = False

        location = numpy.array([0, 0, 0], dtype=numpy.int32)
        for location in self.collision_locations(max_distance):
            x, y, z = location
            cx_ = x >> 4
            cz_ = z >> 4
            if cx is None or cx != cx_ or cz != cz_:
                cx = cx_
                cz = cz_
                try:
                    chunk = self.canvas.world.get_chunk(cx, cz, self.canvas.dimension)
                except ChunkLoadError:
                    chunk = None

            if (
                chunk is not None
                and chunk.block_palette[chunk.blocks[x % 16, y, z % 16]]
                not in UniversalAirLikeBlocks
            ):
                # the block is not air
                if in_air:  # if we have previously found an air block
                    return location, True
            elif not in_air:
                in_air = True
        return location, False

    def get_2d_mouse_location(self) -> Tuple[float, float]:
        """Get the x and z location of the cursor when in 2D mode."""
        x, _, z = self.canvas.camera.location
        width, height = self.canvas.GetSize()
        width = max(1, width)
        height = max(1, height)
        z += 2 * self.canvas.camera.fov * self.canvas.mouse.delta_y / height
        x += (
            2
            * self.canvas.camera.fov
            * self.canvas.camera.aspect_ratio
            * self.canvas.mouse.delta_x
            / width
        )
        x, z = numpy.floor([x, z]) + 0.5
        return x, z

    def closest_block_2d(self, min_y: int = 0) -> Tuple[PointCoordinatesNDArray, bool]:
        """Find the first non-air block above a given y axis.
        Note if there are no non-air blocks above the given axis it will return the min_y value

        :param min_y: The minimum y coordinate to look to.
        :return: Tuple[The block coordinate, was a non-air block found in the range]
        """
        x, z = self.get_2d_mouse_location()
        sub_chunk_size = self.canvas.world.sub_chunk_size
        y = min_y
        try:
            chunk = self.canvas.world.get_chunk(
                int(x // sub_chunk_size),
                int(z // sub_chunk_size),
                self.canvas.dimension,
            )
        except ChunkLoadError:
            pass
        else:
            box_max_chunk = int(min_y // sub_chunk_size)
            sub_chunks = sorted(
                [cy for cy in chunk.blocks.sub_chunks if cy >= box_max_chunk],
                reverse=True,
            )
            if sub_chunks:
                dx, dz = (numpy.floor([x, z]) % sub_chunk_size).astype(numpy.int64)
                for sy in sub_chunks:
                    blocks = numpy.invert(
                        numpy.isin(
                            chunk.blocks.get_section(sy)[dx, ::-1, dz],
                            UniversalAirLikeBlocks,
                        )
                    )
                    if numpy.any(blocks):
                        y = (
                            sub_chunk_size
                            - 1
                            - numpy.argmax(blocks)
                            + sy * sub_chunk_size
                        )
                        y = max(min_y, y)
                        return numpy.asarray((x, y, z)), True
        return numpy.asarray((x, y, z)), False

    def distance_block_3d(
        self,
        distance: float,
        start_location: Optional[numpy.ndarray] = None,
        look_vector: Optional[numpy.ndarray] = None,
    ) -> PointCoordinatesNDArray:
        """The first block location along the camera's look vector that is further away than `distance`.
        :param distance: The distance between the block and the camera.
        :param start_location: The start location. Defaults to the camera location.
        :param look_vector: The unit look vector. Defaults to the camera look vector.
        :return: (x, y, z) numpy array, selection box index
        """
        if look_vector is None:
            look_vector = self.look_vector()
        if start_location is None:
            start_location = self.canvas.camera.location
        position = numpy.array(start_location, dtype=int) + numpy.floor(
            look_vector * distance
        ).astype(int)
        return position

    def collision_locations(
        self,
        max_distance: float = 100,
        start_location: Optional[numpy.ndarray] = None,
        look_vector: Optional[numpy.ndarray] = None,
    ) -> Generator[numpy.ndarray, None, None]:
        """
        The block locations that the camera's look vector passes through.
        :param max_distance: The maximum distance along the look vector to traverse.
        :param start_location: The start location. Defaults to the camera location.
        :param look_vector: The unit look vector. Defaults to the camera look vector.
        :return: A generator of (x, y, z) numpy arrays
        """
        if look_vector is None:
            look_vector = self.look_vector()
        if start_location is None:
            start_location = self.canvas.camera.location
        dx, dy, dz = look_vector

        vectors = numpy.array(
            [look_vector / abs(dx), look_vector / abs(dy), look_vector / abs(dz)]
        )

        locations: List[numpy.ndarray] = []
        # the location within the block
        start: numpy.ndarray = numpy.array(start_location, numpy.float32) % 1

        for axis in range(3):
            vector = vectors[axis]
            vector_size = numpy.sum(vector ** 2) ** 0.5
            # move the start position to the block edge behind the vector
            if vector[axis] > 0:
                initial_offset = -vector * start[axis]
            else:
                initial_offset = -vector * (1 - start[axis])

            block_count = (
                max_distance + numpy.sum(initial_offset ** 2) ** 0.5 - 0.001
            ) // vector_size

            if block_count:
                blocks = numpy.arange(block_count + 1).reshape(-1, 1)
                offsets = blocks * vector + initial_offset
                locations_ = numpy.floor(offsets + start).astype(int)
                if vector[axis] < 0:
                    locations_[:, axis] -= 1

                locations.append(locations_)

        if locations:
            collision_locations = numpy.floor(start_location).astype(int) + numpy.array(
                sorted(
                    numpy.concatenate(locations),
                    key=lambda loc: sum(abs(loc_) for loc_ in loc),
                )
            )
        else:
            collision_locations = start.astype(int).reshape(1, 3)

        for location in collision_locations:
            yield location



from typing import TYPE_CHECKING
from ..edit_canvas_container import EditCanvasContainer

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class BaseBehaviour(EditCanvasContainer):
    """This class implements how the program actually works.
    It detects various inputs and sets the state of the program accordingly.
    This is just a shell of the class to implement the API."""

    def __init__(self, canvas: "EditCanvas"):
        super().__init__(canvas)

    def bind_events(self):
        """Set up all events required to run."""
        pass



from typing import Callable, Optional, Type, TYPE_CHECKING
import inspect
import os
import struct
import hashlib
import wx

from .base_operation_loader import BaseOperationLoader

from amulet_map_editor import log
from amulet.api.level import BaseLevel
from amulet_map_editor.programs.edit.api.operations import (
    OperationUI,
    OperationUIType,
    FixedFunctionUI,
)

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas
    from amulet.api.level import BaseLevel

ValidChrs = set("-_.() abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789")


class UIOperationLoader(BaseOperationLoader):
    def __init__(self, identifier: str, export_dict: dict):
        self._ui: Optional[
            Callable[[wx.Window, "EditCanvas", "BaseLevel"], OperationUI]
        ] = None
        super().__init__(identifier, export_dict)

    def _setup(self, export_dict: dict):
        """Parse the export dictionary and setup as required."""
        options_path = os.path.abspath(
            os.path.join(
                "config",
                "edit_plugins",
                f"""{''.join(c for c in self._name if c in ValidChrs)}_{
                struct.unpack(
                    "H",
                    hashlib.sha1(
                        self.identifier.encode('utf-8')
                    ).digest()[:2]
                )[0]
                }.config""",  # generate a file name that identifiable to the operation but "unique" to the path
            )
        )

        if "operation" in export_dict:
            if inspect.isclass(export_dict["operation"]) and issubclass(
                export_dict["operation"], (wx.Window, wx.Sizer)
            ):
                operation_ui: Type[OperationUI] = export_dict.get("operation", None)
                if issubclass(operation_ui, OperationUI):
                    self._ui = lambda parent, canvas, world: operation_ui(
                        parent, canvas, world, options_path
                    )
                    self._is_valid = True
                else:
                    log.error(
                        f'"operation" must be a subclass of edit.plugins.OperationUI. {self.identifier}'
                    )

            elif callable(export_dict["operation"]):
                operation = export_dict["operation"]
                if operation.__code__.co_argcount == 4:
                    options = export_dict.get("options", {})
                    if isinstance(options, dict):
                        self._ui = lambda parent, canvas, world: FixedFunctionUI(
                            parent, canvas, world, options_path, operation, options
                        )
                        self._is_valid = True
                    else:
                        log.error(
                            f'"operation" in export must be a dictionary if defined. {self.identifier}'
                        )
                else:
                    log.error(
                        f'"operation" function in export must have 4 inputs. {self.identifier}'
                    )
            else:
                log.error(
                    f'"operation" in export must be a callable, or a subclass of wx.Window or wx.Sizer. {self.identifier}'
                )
        else:
            log.error(f'"operation" is not present in export. {self.identifier}')

    def __call__(
        self, parent: wx.Window, canvas: "EditCanvas", world: "BaseLevel"
    ) -> OperationUIType:
        return self._ui(parent, canvas, world)
